import gi.*;

class PureLisp extends SLR1_Grammar {

	PureLisp() throws Exception {

		/**
		* Lexical Specification
		*/
		put("IDENTIFIER", Lexicon.expression("[[:alnum:]]+"));
		put("SPACE", Lexicon.expression("[[:space:]]+"));
		put("COMMENT", Lexicon.expression("\\{[^}]*\\}"));
		put("T", Lexicon.expression("T|t"));
		put("NIL", Lexicon.expression("NIL|nil"));

		/**
		* Semantic Specification
		*/
		Semantics interpret = new Semantics() {
			public void evaluate(ParseTree tree) throws Exception {
				System.out.println(output(value(tree.phrase[0].attribute, null)));
			}
		};
		Semantics identity = new Semantics() {
			public void evaluate(ParseTree tree) {
				tree.attribute = tree.phrase[0].attribute;
			}
		};
		Semantics truth = new Semantics() {
			public void evaluate(ParseTree tree) {
				tree.attribute = "T";
			}
		};
		Semantics list_zero = new Semantics() {
			public void evaluate(ParseTree tree) {
				tree.attribute = null;
			}
		};
		Semantics list_zero_plus = new Semantics() {
			public void evaluate(ParseTree tree) {
				tree.attribute = tree.phrase[1].attribute;
			}
		};
		Semantics list_one_plus = new Semantics() {
			public void evaluate(ParseTree tree) {
				ParseTree[] phrase = tree.phrase;
				tree.attribute = cons(phrase[1].attribute, phrase[2].attribute);
			}
		};
		Semantics list_two = new Semantics() {
			public void evaluate(ParseTree tree) {
				ParseTree[] phrase = tree.phrase;
				tree.attribute = cons(phrase[1].attribute, cons(phrase[2].attribute, null));
			}
		};
		Semantics list_three = new Semantics() {
			public void evaluate(ParseTree tree) {
				ParseTree[] phrase = tree.phrase;
				tree.attribute = cons(phrase[1].attribute, cons(phrase[2].attribute, cons(phrase[3].attribute, null)));
			}
		};
		Semantics list_prepend = new Semantics() {
			public void evaluate(ParseTree tree) {
				ParseTree[] phrase = tree.phrase;
				tree.attribute = cons(phrase[0].attribute, phrase[1].attribute);
			}
		};
		Semantics function = new Semantics() {
			public void evaluate(ParseTree tree) {
				ParseTree[] phrase = tree.phrase;
				tree.attribute = cons(phrase[1].attribute, cons(phrase[3].attribute, cons(phrase[5].attribute, null)));
			}
		};

		/**
		* Syntax Specification
		*/
		put("Start", new Object[][] {
			{"Expression", interpret}
		});
		put("Expression", new Object[][] {
			{"Atom", identity},
			{"List", identity}
		});
		put("Atom", new Object[][] {
			{"T", truth},
			{"IDENTIFIER", identity}
		});
		put("List", new Object[][] {
			{"NIL", list_zero},
			{"(", "quote", "Expression", ")", list_two},
			{"(", "car", "Expression", ")", list_two},
			{"(", "cdr", "Expression", ")", list_two},
			{"(", "cons", "Expression", "Expression", ")", list_three},
			{"(", "eq", "Expression", "Expression", ")", list_three},
			{"(", "atom", "Expression", ")", list_two},
			{"(", "cond", "Cases", ")", list_one_plus},
			{"(", "Expressions", ")", list_zero_plus},
			{"(", "AnonymousFunction", "Expressions", ")", list_one_plus},
			{"(", "NamedFunction", "Expressions", ")", list_one_plus}
		});
		put("Cases", new Object[][] {
			{"Case", "Cases", list_prepend},
			{list_zero}
		});
		put("Expressions", new Object[][] {
			{"Expression", "Expressions", list_prepend},
			{list_zero}
		});
		put("AnonymousFunction", new Object[][] {
			{"(", "lambda", "(", "Formals", ")", "Expression", ")", function}
		});
		put("NamedFunction", new Object[][] {
			{"(", "label", "IDENTIFIER", "AnonymousFunction", ")", list_three}
		});
		put("Case", new Object[][] {
			{"(", "Expression", "Expression", ")", list_two}
		});
		put("Formals", new Object[][] {
			{"IDENTIFIER", "Formals", list_prepend},
			{list_zero}
		});
	}

	class Pair {

		Object car;
		Object cdr;

		Pair(Object car, Object cdr) {
			this.car = car;
			this.cdr = cdr;
		}

		public String toString() {
			try {
				return output(this);
			}
			catch (Exception error) {
				return null;
			}
		}
	}

	Object car(Object t) throws Exception {
		if (t instanceof Pair) return ((Pair)t).car;
		else throw new Exception("invalid car operation");
	}

	Object cdr(Object t) throws Exception {
		if (t instanceof Pair) return ((Pair)t).cdr;
		else throw new Exception("invalid cdr operation");
	}

	Object cadr(Object t) throws Exception {
		return car(cdr(t));
	}

	Object caddr(Object t) throws Exception {
		return car(cdr(cdr(t)));
	}

	Object caar(Object t) throws Exception {
		return car(car(t));
	}

	Object cadar(Object t) throws Exception {
		return car(cdr(car(t)));
	}

	Object caddar(Object t) throws Exception {
		return car(cdr(cdr(car(t))));
	}

	Object cons(Object s, Object t) {
		return new Pair(s, t);
	}

	boolean eq(Object s, Object t) throws Exception {
		if (s == null && t == null)
			return true;
		else if (s instanceof String && t instanceof String)
			return s.equals(t);
		else if (s instanceof Pair && t instanceof Pair)
			return eq(car(s), car(t)) && eq(cdr(s), cdr(t));
		else
			return false;
	}

	Object caseValue(Object cases, Object global) throws Exception {
		if (cases == null)
			throw new Exception("invalid empty case expression");
		else if (value(caar(cases), global) != null)
			return value(cadar(cases), global);
		else
			return caseValue(cdr(cases), global);
	}

	Object binding(Object identifier, Object global) throws Exception {
		if (global == null)
			throw new Exception("identifier " + identifier + " not defined");
		else if (eq(identifier, caar(global)))
			return car(global);
		else
			return binding(identifier, cdr(global));
	}

	Object append(Object local, Object global) throws Exception {
		if (local == null)
			return global;
		else
			return cons(car(local), append(cdr(local), global));
	}

	Object local(Object formals, Object values) throws Exception {
		if (formals == null && values == null)
			return null;
		else
			return cons(cons(car(formals), car(values)), local(cdr(formals), cdr(values)));
	}

	Object values(Object actuals, Object global) throws Exception {
		if (actuals == null)
			return null;
		else
			return cons(value(car(actuals), global), values(cdr(actuals), global));
	}

	Object value(Object t, Object global) throws Exception {
		if (t == null)
			return t;
		else if (t instanceof String)
			if (t.equals("T"))
				return t;
			else
				return cdr(binding(t, global));
		else if (car(t) instanceof String)
			if (car(t).equals("quote"))
				return cadr(t);
			else if (car(t).equals("car"))
				return car(value(cadr(t), global));
			else if (car(t).equals("cdr"))
				return cdr(value(cadr(t), global));
			else if (car(t).equals("cons"))
				return cons(value(cadr(t), global), value(caddr(t), global));
			else if (car(t).equals("eq"))
				if (eq(value(cadr(t), global), value(caddr(t), global)))
					return "T";
				else
					return null;
			else if (car(t).equals("atom"))
				if (value(cadr(t), global) instanceof String)
					return "T";
				else
					return null;
			else if (car(t).equals("cond"))
				return caseValue(cdr(t), global);
			else
				return value(cons(cdr(binding(car(t), global)), cdr(t)), global);
		else if (caar(t) instanceof String)
			if (caar(t).equals("lambda"))
				return value(caddar(t), append(local(cadar(t), values(cdr(t), global)), global));
			else if (caar(t).equals("label"))
				return value(cons(caddar(t), cdr(t)), cons(cons(cadar(t), caddar(t)), global));
			else
				throw new Exception("invalid doubly nested operation");
		else
			throw new Exception("invalid singly nested operation");
	}

	String outputCdr(Object t) throws Exception {
		if (t == null)
			return "";
		else if (t instanceof String)
			return "." + t;
		else
			return " " + output(car(t)) + outputCdr(cdr(t));
	}

	String output(Object t) throws Exception {
		if (t == null)
			return "NIL";
		else if (t instanceof String)
			return (String)t;
		else
			return "(" + output(car(t)) + outputCdr(cdr(t)) + ")";
	}

	public static void main(String[] argv) throws Exception {
		new PureLisp().interpret(argv);
	}
}
